@@ -25,9 +25,12 @@
 
 #include "config.h"
 
+#define DEBUG_FLAG GKR_DEBUG_OPERATION
+#include "gkr-debug.h"
 #include "gkr-misc.h"
 #include "gkr-operation.h"
 #include "gkr-session.h"
+
 #include "gnome-keyring.h"
 #include "gnome-keyring-private.h"
 
@@ -45,13 +48,19 @@ struct _GkrOperation {
 	gint refs;
 	gint result;
 
+	/* Call information */
 	DBusConnection *conn;
-	DBusPendingCall *pending;
 	gboolean prompting;
+	DBusMessage *request;
+
+	/* Asynchronous calls */
+	gboolean asynchronous;
+	DBusPendingCall *pending;
 
 	/* Some strange status fields */
 	gboolean was_keyring;
 
+	/* Callback information */
 	GQueue callbacks;
 	GSList *completed;
 };
@@ -68,7 +77,10 @@ static void
 operation_clear_callbacks (GkrOperation *op)
 {
 	GSList *l;
+
 	g_assert (op);
+	gkr_debug ("%p", op);
+
 	while (!g_queue_is_empty (&op->callbacks))
 		gkr_callback_free (g_queue_pop_head (&op->callbacks));
 	g_queue_clear (&op->callbacks);
@@ -88,6 +100,7 @@ operation_unref (gpointer data)
 		return FALSE;
 
 	if (op->pending) {
+		gkr_debug ("%p: cancelling: %p", op, op->pending);
 		dbus_pending_call_cancel (op->pending);
 		dbus_pending_call_unref (op->pending);
 		op->pending = NULL;
@@ -95,6 +108,8 @@ operation_unref (gpointer data)
 
 	operation_clear_callbacks (op);
 
+	gkr_debug ("%p: freeing", op);
+
 	if (op->conn) {
 		dbus_connection_unref (op->conn);
 		op->conn = NULL;
@@ -112,19 +127,6 @@ gkr_operation_unref (gpointer data)
 	operation_unref (op);
 }
 
-gpointer
-gkr_operation_pending_and_unref (GkrOperation *op)
-{
-	/* Return op, if not the last reference */
-	if (!operation_unref (op))
-		return op;
-
-	/* Not sure what to do here, but at least we can print a message */
-	g_message ("an libgnome-keyring operation completed unsafely before "
-	           "the function starting the operation returned.");
-	return NULL;
-}
-
 GnomeKeyringResult
 gkr_operation_unref_get_result (GkrOperation *op)
 {
@@ -140,6 +142,7 @@ gkr_operation_new (gpointer callback, GkrCallbackType callback_type,
 	GkrOperation *op;
 
 	op = g_slice_new0 (GkrOperation);
+	gkr_debug ("%p", op);
 
 	op->refs = 1;
 	op->result = INCOMPLETE;
@@ -201,6 +204,8 @@ on_complete (GkrOperation *op)
 	cb = g_queue_pop_tail (&op->callbacks);
 	g_assert (cb);
 
+	gkr_debug ("%p", op);
+
 	/* Free all the other callbacks */
 	operation_clear_callbacks (op);
 
@@ -232,9 +237,11 @@ void
 gkr_operation_complete_later (GkrOperation *op, GnomeKeyringResult res)
 {
 	g_return_if_fail (op);
-	if (gkr_operation_set_result (op, res))
+	if (gkr_operation_set_result (op, res)) {
+		gkr_debug ("%p", op);
 		g_idle_add_full (G_PRIORITY_DEFAULT_IDLE, on_complete_later,
 		                 gkr_operation_ref (op), gkr_operation_unref);
+	}
 }
 
 static DBusHandlerResult
@@ -255,6 +262,7 @@ on_name_changed_filter (DBusConnection *connection, DBusMessage *message, void *
 		    new_owner && g_str_equal ("", new_owner)) {
 
 			/* Clear any session, when the service goes away */
+			gkr_debug ("secret service went away");
 			gkr_session_clear ();
 		}
 
@@ -277,8 +285,10 @@ connect_to_service (void)
 	 */
 
 	if (!dbus_connection) {
-		if (!g_getenv ("DBUS_SESSION_BUS_ADDRESS"))
+		if (!g_getenv ("DBUS_SESSION_BUS_ADDRESS")) {
+			gkr_debug ("no DBUS_SESSION_BUS_ADDRESS var set");
 			return NULL;
+		}
 
 		conn = dbus_bus_get_private (DBUS_BUS_SESSION, &derr);
 		if (conn == NULL) {
@@ -302,9 +312,11 @@ connect_to_service (void)
 		{
 			if (dbus_connection) {
 				dbus_connection_unref (conn);
+				gkr_debug ("race. already have a dbus connection");
 			} else {
 				egg_dbus_connect_with_mainloop (conn, NULL);
 				dbus_connection = conn;
+				gkr_debug ("created and initialized dbus connection");
 			}
 		}
 		G_UNLOCK (dbus_connection);
@@ -339,6 +351,7 @@ on_pending_call_notify (DBusPendingCall *pending, void *user_data)
 	GkrOperation *op = user_data;
 	DBusMessage *reply;
 
+	gkr_debug ("%p: notified: %p", op, pending);
 	g_assert (pending == op->pending);
 
 	reply = dbus_pending_call_steal_reply (pending);
@@ -355,15 +368,16 @@ on_pending_call_notify (DBusPendingCall *pending, void *user_data)
 	gkr_operation_unref (op);
 }
 
-void
-gkr_operation_request (GkrOperation *op, DBusMessage *req)
+static void
+send_with_pending (GkrOperation *op)
 {
 	int timeout = -1;
 
-	g_return_if_fail (req);
 	g_assert (op);
+	g_assert (op->request);
+	g_assert (!op->pending);
 
-#if WITH_TESTS
+#if WITH_TESTABLE
 	timeout = INT_MAX;
 #endif
 
@@ -371,14 +385,16 @@ gkr_operation_request (GkrOperation *op, DBusMessage *req)
 		op->conn = connect_to_service ();
 
 	if (op->conn) {
-		g_assert (!op->pending);
-		if (!dbus_connection_send_with_reply (op->conn, req, &op->pending, timeout))
+		gkr_debug ("%p: sending request", op);
+		if (!dbus_connection_send_with_reply (op->conn, op->request,
+		                                      &op->pending, timeout))
 			g_return_if_reached ();
+		dbus_message_unref (op->request);
+		op->request = NULL;
 	}
 
 	if (op->pending) {
-		if (gkr_decode_is_keyring (dbus_message_get_path (req)))
-			gkr_operation_set_keyring_hint (op);
+		gkr_debug ("%p: has pending: %p", op, op->pending);
 		dbus_pending_call_set_notify (op->pending, on_pending_call_notify,
 		                              gkr_operation_ref (op), gkr_operation_unref);
 	} else {
@@ -386,31 +402,130 @@ gkr_operation_request (GkrOperation *op, DBusMessage *req)
 	}
 }
 
+void
+gkr_operation_request (GkrOperation *op, DBusMessage *req)
+{
+	g_return_if_fail (req);
+
+	g_assert (op);
+	g_assert (op->request == NULL);
+	g_assert (op->pending == NULL);
+
+	op->request = dbus_message_ref (req);
+
+	if (gkr_decode_is_keyring (dbus_message_get_path (req)))
+		gkr_operation_set_keyring_hint (op);
+
+	/*
+	 * The first requests waits until we know if we're asynchronous or not.
+	 * Thereafter all async requests are sent directly from here. See
+	 * gkr_operation_pending_and_unref() and gkr_operation_block_and_unref().
+	 */
+
+	if (op->asynchronous)
+		send_with_pending (op);
+}
+
+gpointer
+gkr_operation_pending_and_unref (GkrOperation *op)
+{
+	g_assert (op);
+
+	/*
+	 * Only the first asynch requests waits until here to be sent. Further
+	 * requests on this operation get sent from gkr_operation_request() directly.
+	 */
+
+	g_assert (!op->asynchronous);
+	op->asynchronous = TRUE;
+
+	if (op->request)
+		send_with_pending (op);
+
+	/* Return op, if not the last reference */
+	if (!operation_unref (op))
+		return op;
+
+	/* Not sure what to do here, but at least we can print a message */
+	g_message ("a libgnome-keyring operation completed unsafely before "
+	           "the function starting the operation returned.");
+	return NULL;
+}
+
+static DBusMessage*
+send_with_reply_and_block (DBusConnection *conn, DBusMessage *message,
+                           int timeout)
+{
+	DBusMessage *reply;
+	DBusPendingCall *pending;
+
+	g_assert (conn);
+	g_assert (message);
+
+	if (!dbus_connection_send_with_reply (conn, message, &pending, timeout))
+		g_return_val_if_reached (NULL);
+
+	if (pending == NULL) {
+		gkr_debug ("couldn't send message");
+		return NULL;
+	}
+
+	dbus_pending_call_block (pending);
+	reply = dbus_pending_call_steal_reply (pending);
+	dbus_pending_call_unref (pending);
+
+	return reply;
+}
+
 GnomeKeyringResult
 gkr_operation_block_and_unref (GkrOperation *op)
 {
-	DBusPendingCall *pending;
+	DBusMessage *reply = NULL;
+	int timeout = -1;
+
 	g_return_val_if_fail (op, BROKEN);
 
+#if WITH_TESTABLE
+	timeout = INT_MAX;
+#endif
+
+	gkr_debug ("%p: processing", op);
+
+	g_assert (!op->pending);
+	g_assert (!op->asynchronous);
+
 	while ((int) gkr_operation_get_result (op) == INCOMPLETE) {
-		if (op->pending) {
-			/*
-			 * DBus has strange behavior that can complete a pending call
-			 * in another thread and somehow does this without calling our
-			 * on_pending_call_notify. So guard against this brokenness.
-			 */
-			pending = op->pending;
-			dbus_pending_call_block (pending);
-			if (op->pending == pending) {
-				g_return_val_if_fail (dbus_pending_call_get_completed (pending), BROKEN);
-				on_pending_call_notify (pending, op);
+
+		/* No connection then connect */
+		if (!op->conn) {
+			op->conn = connect_to_service ();
+			if (!op->conn)
+				gkr_operation_complete (op, GNOME_KEYRING_RESULT_IO_ERROR);
+
+		/* A dbus request, then send and wait for reply */
+		} else if (op->request) {
+			gkr_debug ("%p: blocking request", op);
+			reply = send_with_reply_and_block (op->conn, op->request, timeout);
+			dbus_message_unref (op->request);
+			op->request = NULL;
+
+			if (reply != NULL) {
+				callback_with_message (op, reply);
+				dbus_message_unref (reply);
+			} else {
+				gkr_operation_complete (op, GNOME_KEYRING_RESULT_IO_ERROR);
 			}
+
+		/* Prompting then wait until prompting done */
 		} else if (op->prompting) {
 			dbus_connection_flush (op->conn);
+			gkr_debug ("%p: blocking on prompt", op);
 			while (op->prompting && (int) gkr_operation_get_result (op) == INCOMPLETE) {
 				if (!dbus_connection_read_write_dispatch (op->conn, 200))
 					break;
 			}
+
+		/* Not prompting and not requesting == bug in the code */
 		} else {
 			g_assert_not_reached ();
 		}
@@ -420,6 +535,8 @@ gkr_operation_block_and_unref (GkrOperation *op)
 	if (!g_queue_is_empty (&op->callbacks))
 		on_complete (op);
 
+	gkr_debug ("%p: done", op);
+
 	return gkr_operation_unref_get_result (op);
 }
 
@@ -444,6 +561,7 @@ gkr_operation_handle_errors (GkrOperation *op, DBusMessage *reply)
 
 	if (dbus_set_error_from_message (&derr, reply)) {
 		if (dbus_error_has_name (&derr, ERROR_NO_SUCH_OBJECT)) {
+			gkr_debug ("%p: no-such-object", op);
 			if (was_keyring)
 				res = GNOME_KEYRING_RESULT_NO_SUCH_KEYRING;
 			else
@@ -470,7 +588,10 @@ static void
 on_prompt_completed (void *user_data)
 {
 	on_prompt_args *args = user_data;
+
 	g_return_if_fail (args->op->prompting);
+	gkr_debug ("%p", args->op);
+
 	gkr_operation_unref (args->op);
 	args->op->prompting = FALSE;
 }
@@ -488,8 +609,10 @@ on_prompt_signal (DBusConnection *connection, DBusMessage *message, void *user_d
 
 	g_assert (args);
 
-	if (!args->path || !args->op->prompting)
+	if (!args->path || !args->op->prompting) {
+		gkr_debug ("%p: received prompt signal while not prompting", args->op);
 		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+	}
 
 	/* org.freedesktop.Secret.Prompt.Completed(BOOLEAN dismissed, VARIANT result) */
 	if (dbus_message_has_path (message, args->path) &&
@@ -506,10 +629,13 @@ on_prompt_signal (DBusConnection *connection, DBusMessage *message, void *user_d
 
 		op = gkr_operation_ref (args->op);
 
-		if (dismissed)
+		if (dismissed) {
+			gkr_debug ("%p: prompt was dismissed", op);
 			gkr_operation_complete (op, GNOME_KEYRING_RESULT_CANCELLED);
-		else
+		} else {
+			gkr_debug ("%p: prompt was completed", op);
 			callback_with_message (op, message);
+		}
 
 		if (op->prompting)
 			dbus_connection_remove_filter (args->op->conn, on_prompt_signal, args);
@@ -528,7 +654,7 @@ on_prompt_signal (DBusConnection *connection, DBusMessage *message, void *user_d
 		if (object_name && g_str_equal (gkr_service_name (), object_name) &&
 		    new_owner && g_str_equal ("", new_owner)) {
 
-			g_message ("Secret service disappeared while waiting for prompt");
+			g_message ("secret service disappeared while waiting for prompt");
 			op = gkr_operation_ref (args->op);
 			gkr_operation_complete (op, GNOME_KEYRING_RESULT_IO_ERROR);
 			if (op->prompting)
@@ -591,6 +717,8 @@ gkr_operation_prompt (GkrOperation *op, const gchar *prompt)
 	window_id = "";
 	dbus_message_append_args (req, DBUS_TYPE_STRING, &window_id, DBUS_TYPE_INVALID);
 
+	gkr_debug ("%p: calling prompt method", op);
+
 	gkr_operation_push (op, on_prompt_result, GKR_CALLBACK_OP_MSG, args, on_prompt_free);
 	gkr_operation_request (op, req);
 	dbus_message_unref (req);
